from __future__ import annotations

from typing import TYPE_CHECKING, List
from enum import Enum

from .integration import ServerIntegration

if TYPE_CHECKING:
    from .user import User
    from .types import (
        IntegrationType as ConnectionType,
        Connection as ConnectionPayload,
    )

__all__ = ("Connection",)


class Visibility(Enum):
    none = 0
    everyone = 1


class Connection:
    """Represents connection object that the user has attached.

    Attributes
    -----------
    id: :class:`int`
        The connection's unique id.
    name: :class:`str`
        The connection's name.
    type: :class:`str`
        The service of the connection (twitch, youtube).
    revoked: :class:`bool`
       If the connection is revoked.
    verified: :class:`bool`
        If the connection is verified.
    friend_sync: :class:`bool`
        If friend sync is enabled for this connection.
    show_activity: :class:`bool`
        If activities related to this connection
        will be shown in presence updates
    visibility: :class:`Visibility`
        Visibility of this connection
    """

    __slots__ = (
        "id",
        "name",
        "type",
        "revoked",
        "integrations",
        "verified",
        "friend_sync",
        "show_activity",
        "visibility",
        "user",
    )

    def __init__(self, data: ConnectionPayload, user: User) -> None:
        self.user: User = user
        self.id = data["id"]
        self.name = data["name"]
        self.type: ConnectionType = data["type"]
        self.revoked: bool = data.get("revoked", False)
        self.integrations: List[ServerIntegration] = [
            ServerIntegration(x, user=user) for x in data.get("integrations", [])
        ]
        self.verified: bool = data["verified"]
        self.friend_sync: bool = data["friend_sync"]
        self.show_activity: bool = data["show_activity"]
        self.visibility: Visibility = Visibility(data["visibility"])

    def __repr__(self) -> str:
        return f"<Connection id={self.id} name={self.name!r} type={self.type!r}>"

    def __str__(self) -> str:
        return self.name



from __future__ import annotations

from typing import TYPE_CHECKING
from datetime import datetime, timedelta

if TYPE_CHECKING:
    from .types import AccessToken as AccessTokenPayload

    from typing_extensions import Self

__all__ = ("AccessToken",)


class AccessToken:
    """Represents a access token response information provided by discord.

    Attributes
    ----------
    access_token: :class:`str`
        The user's access token.
    token_type: :class:`str`
        The token type.
    refresh_token: :class:`str`
        The user's refresh token.
    expires_in: :class:`int`
        When this access token will expire (in seconds).
    """

    def __init__(self, data: AccessTokenPayload) -> None:
        self.access_token: str = data["access_token"]
        self.token_type: str = data["token_type"]
        self.refresh_token: str = data["refresh_token"]
        self.scope: str = data["scope"]
        self.expires_in: int = data["expires_in"]

    @property
    def expires_at(self) -> datetime:
        return datetime.now() - timedelta(seconds=self.expires_in)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from .types import (
        IntegrationApp as IntegrationAppPayload,
    )

__all__ = ("IntegrationApp",)


class IntegrationApp:
    """Represents the integration application provided by Discord.

    Attributes
    ----------
    id: :class:`int`
        The application unique id.
    name: :class:`str`
        The application name.
    description: :class:`str`
        The description of the app.
    """

    __slots__ = (
        "id",
        "name",
        "_icon_hash",
        "description",
        "bot",
    )

    def __init__(self, data: IntegrationAppPayload) -> None:
        self.id: int = int(data["id"])
        self.name: str = data["name"]
        self._icon_hash: Optional[str] = data.get("icon")
        self.description: str = data["description"]

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} name={self.name!r} description={self.description!r}"

    @property
    def icon(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns the app's icon url, if available."""
        if not self._icon_hash:
            return None
        return f"https://cdn.discordapp.com/app-icons/{self.id}/{self._icon_hash}.png?size=1024"



from __future__ import annotations

from typing import Optional, TYPE_CHECKING

from .utils import parse_time
from .flags import Permissions

if TYPE_CHECKING:
    from datetime import datetime

    from .user import User
    from .guild import Guild
    from .types import Member as MemberPayload, SnowflakeList


__all__ = ("Member",)


class Member:
    """Represents a Discord member to a :class:`Guild`.

    .. container:: operations

        ..  describe:: x == y

            Compare the ids of the two members and verify that they are the same.

        .. describe:: x != y

            Compare the ids of the two members and verify that they are the different.

        .. describe:: hash(x)

            Return the member's hash.

        .. describe:: str(x)

            Returns the member's id with discriminator.

    Attributes
    ----------
    user: :class:`User`
        The user representation of this member.
    guild: :class:`Guild`
        The guild that the user is member of.
    nick: :class:`str`
        The guild specific nickname of the user.
    roles: :class:`SnowflakeList`
        The list of role ids that the member has
    joined_at: :class:`datetime.datetime`
        An aware datetime object that specifies the date and time in UTC that the member joined the guild.
    premium_since: Optional[:class:`datetime.datetime`]
        An aware datetime object that specifies the date andtime in UTC that the member "Nitro boost" the guild.
        This could be ``None``.
    deaf: :class:`bool`
        If the user is deafened in voice channels.
    mute: :class:`bool`
        If the user is muted in voice channels
    pending: :class:`bool`
        If the user has not yet passed the guild's Membership Screening requirements.
    timed_out_until: Optional[:class:`datetime.datetime`]
        An aware datetime object that specifies the date and time in UTC that the user's time out will expire.
        This is ``None`` when the user is not timed out.
    """

    __slots__ = (
        "user",
        "guild",
        "nick",
        "_avatar_hash",
        "roles",
        "joined_at",
        "premium_since",
        "deaf",
        "mute",
        "pending",
        "_permissions",
        "timed_out_until",
    )

    def __init__(self, data: MemberPayload, user: User, guild: Guild) -> None:
        self.user: User = user

        try:
            self.user._update(data["user"])
        except KeyError:
            pass

        self.guild: Guild = guild

        self.nick: Optional[str] = data.get("nick")
        self._avatar_hash: Optional[str] = data.get("avatar")
        self.roles: SnowflakeList = data["roles"]
        self.joined_at: datetime = parse_time(data["joined_at"])
        self.premium_since: Optional[datetime] = parse_time(data.get("premium_since"))
        self.deaf: bool = data["deaf"]
        self.mute: bool = data["mute"]
        self.pending: Optional[bool] = data.get("pending")
        self._permissions: int = int(data.get("permissions", 0))
        self.timed_out_until: Optional[datetime] = parse_time(
            data.get("communication_disabled_until")
        )

    def __str__(self) -> str:
        return str(self.user)

    def __repr__(self) -> str:
        return (
            f"<Member id={self.user.id} name={self.user.name!r} discriminator={self.user.discriminator!r}"
            f" bot={self.user.bot} nick={self.nick!r} guild={self.guild!r}>"
        )

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Member) and other.user.id == self.user.id

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.user)

    @property
    def permissions(self) -> Permissions:
        """:class:`Permissions`: Returns the resolved permissions that the oauth2 user has in this guild."""
        return Permissions(self._permissions)

    @property
    def mention(self) -> str:
        """:class:`str`: Returns a string format to mention this member in discord."""
        return f"<@{self.user.id}>"

    @property
    def display_name(self) -> str:
        """:class:`str`: Returns the user's display name."""
        return self.nick or self.user.name

    @property
    def guild_avatar_url(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns the link of cdn for the avatar the member has."""
        if self._avatar_hash is None:
            return None

        animated = self._avatar_hash.startswith("a_")
        fmt = "gif" if animated else "png"
        return f"https://cdn.discordapp.com/guilds/{self.guild.id}/users/{self.user.id}/avatars/{self._avatar_hash}.{fmt}?size=1024"

    @property
    def display_avatar(self) -> str:
        """:class:`str`: Returns the member's display avatar."""
        return self.guild_avatar_url or self.user.display_avatar



from __future__ import annotations

from typing import (
    Any,
    Dict,
    Optional,
    Union,
    overload,
    Type,
    TypeVar,
    Any,
    TYPE_CHECKING,
)
from enum import Enum

import datetime
import json

try:
    import orjson
except ModuleNotFoundError:
    ORJSON = False
else:
    ORJSON = True

if TYPE_CHECKING:
    from aiohttp import ClientResponse

__all__ = ("_JSON_LOADER", "MISSING", "parse_time", "json_or_text")

E = TypeVar("E", bound="Enum")
T = TypeVar("T")

if ORJSON:

    def _to_json(obj: Any) -> str:
        return orjson.dumps(obj).decode("utf-8")  # type: ignore

    _JSON_LOADER = orjson.loads  # type: ignore
else:

    def _to_json(obj: Any) -> str:
        return json.dumps(obj, separators=(",", ":"), ensure_ascii=True)

    _JSON_LOADER = json.loads


class _Missing:
    __slots__ = ()

    def __eq__(self, other):
        return False

    def __bool__(self):
        return False

    def __hash__(self):
        return 0

    def __repr__(self):
        return "..."


MISSING: Any = _Missing()


@overload
def parse_time(timestamp: str) -> datetime.datetime:
    ...


@overload
def parse_time(timestamp: None) -> None:
    ...


def parse_time(timestamp: Optional[str]) -> Optional[datetime.datetime]:
    if not timestamp:
        return None
    return datetime.datetime.fromisoformat(timestamp)


@overload
def try_enum(cls: Type[E], value: None) -> None:
    ...


@overload
def try_enum(cls: Type[E], value: int) -> E:
    ...


def try_enum(cls: Type[E], value: Optional[int]) -> Optional[E]:
    try:
        return cls(value)
    except (ValueError, TypeError):
        return None


async def json_or_text(resp: ClientResponse) -> Union[Dict[str, Any], str]:
    text = await resp.text(encoding="utf-8")
    try:
        if resp.headers["content-type"] == "application/json":
            return _JSON_LOADER(text)
    except KeyError:
        pass

    return text


def copy_doc(target: Any):
    def decorator(overridden: Any):
        overridden.__doc__ = target.__doc__
        return overridden

    return decorator


# This is to avoid pyright raise a error when excpeted Snowflake and int received,
# this function will ensure that the Snowflake is a int, if available.
def get_snowflake(data: Any, key: str) -> Optional[int]:
    try:
        value = data[key]
    except KeyError:
        return None
    else:
        return value and int(value)



__title__ = "disco.oauth2"
__author__ = "martimartins"
__author_email__ = "martim13artins13@gmail.com"
__license__ = "MIT"
__copyright__ = "Copyright 2022-present Martim Martins"
__version__ = "1.0a"

from typing import Literal, NamedTuple

from .client import *
from .errors import *
from .flags import *
from .guild import *
from .http import *
from .token import *
from .user import *
from .connection import *
from .integration import *
from .member import *
from .token import *
from .appinfo import *

_VersionInfo = NamedTuple(
    "_VersionInfo",
    major=int,
    minor=int,
    micro=int,
    releaselevel=Literal["alpha", "beta", "candidate", "final"],
    serial=int,
)

version_info = _VersionInfo(major=1, minor=0, micro=0, releaselevel="alpha", serial=0)



from __future__ import annotations

import aiohttp
import sys
import logging
import asyncio

from urllib.parse import quote as _quote

from . import __version__
from .errors import HTTPException, Forbidden, NotFound, DiscordServerError
from .utils import MISSING, json_or_text

from typing import (
    ClassVar,
    Optional,
    TypeVar,
    overload,
    TYPE_CHECKING,
    Coroutine,
    Any,
    List,
)

if TYPE_CHECKING:
    from .types import (
        Snowflake,
        SnowflakeList,
        User,
        AccessToken as AccessTokenResponse,
        Member,
        PartialGuild as PartialGuildPayload,
        Connection as ConnectionPayload,
    )

    T = TypeVar("T")
    Response = Coroutine[Any, Any, T]
    Session = aiohttp.ClientSession

_log = logging.getLogger(__name__)

__all__ = ("AsyncHTTP",)


class Route:
    BASE: ClassVar[str] = "https://discord.com/api/v10"

    def __init__(self, method: str, path: str, **params: Any) -> None:
        self.method = method
        self.url = self.BASE + path

        if params:
            self.url = self.url.format_map(
                {k: _quote(v) if isinstance(v, str) else v for k, v in params.items()}
            )


class BaseHTTP:
    """Represents an HTTP client sending HTTP requests to the Discord API[oauth2]."""

    def __init__(
        self,
        *,
        client_id: int,
        client_secret: str,
        redirect_uri: str,
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
    ) -> None:
        self.client_id = client_id
        self._client_secret = client_secret
        self.redirect_uri = redirect_uri

        # Proxy support
        self.proxy: Optional[str] = proxy
        self.proxy_auth: Optional[aiohttp.BasicAuth] = proxy_auth

    @overload
    def _create_session(self) -> Any:
        ...

    @overload
    def _create_session(self) -> Coroutine[Any, Any, aiohttp.ClientSession]:
        ...

    def _create_session(self) -> Any:
        raise NotImplementedError

    @overload
    def request(self, route: Route, **kwargs) -> Any:
        ...

    @overload
    def request(self, route: Route, **kwargs) -> Response[Any]:
        ...

    def request(self, route: Route, **kwargs) -> Any:
        raise NotImplementedError

    @overload
    def close(self) -> None:
        ...

    @overload
    def close(self) -> Coroutine[Any, Any, None]:
        ...

    def close(self) -> Any:
        raise NotImplementedError

    # Api methods

    def get_current_auth(self) -> Response[Any]:
        r = Route("GET", "/oauth2/@me")
        payload = {
            "client_id": self.client_id,
            "client_secret": self._client_secret,
        }

        return self.request(r, data=payload)

    def get_user(self, access_token: str) -> Response[User]:
        r = Route("GET", "/users/@me")
        headers = {"Authorization": f"Bearer {access_token}"}

        return self.request(r, headers=headers)

    def get_member(self, access_token: str, guild_id: Snowflake) -> Response[Member]:
        r = Route("GET", "/users/@me/guilds/{guild_id}/member", guild_id=guild_id)
        headers = {"Authorization": f"Bearer {access_token}"}

        return self.request(r, headers=headers)

    def add_user_guild(
        self,
        user_id: Snowflake,
        guild_id: Snowflake,
        access_token: str,
        bot_token: str,
        *,
        nick: Optional[str] = None,
        roles: Optional[SnowflakeList] = None,
        mute: Optional[bool] = False,
        deaf: Optional[bool] = False,
    ) -> Response[None]:
        headers = {"Authorization": bot_token}
        payload = {
            "access_token": access_token,
            "mute": mute,
            "deaf": deaf,
        }
        if nick:
            payload["nick"] = nick
        if roles:
            payload["roles"] = nick

        r = Route(
            "PUT",
            "/guilds/{guild_id}/members/{user_id}",
            guild_id=guild_id,
            user_id=user_id,
        )
        return self.request(r, data=payload, headers=headers)

    def get_user_guilds(self, access_token: str) -> Response[List[PartialGuildPayload]]:
        r = Route("GET", "/users/@me/guilds")
        headers = {"Authorization": f"Bearer {access_token}"}

        return self.request(r, headers=headers)

    def get_user_connections(
        self, access_token: str
    ) -> Response[List[ConnectionPayload]]:
        r = Route("GET", "/users/@me/connections")
        headers = {"Authorization": f"Bearer {access_token}"}

        return self.request(r, headers=headers)

    def refresh_token(self, refresh_token: str) -> Response[AccessTokenResponse]:
        r = Route("GET", "/oauth2/token")
        payload = {
            "client_id": self.client_id,
            "client_secret": self._client_secret,
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
        }

        return self.request(r, data=payload)

    def exchange_code(
        self, code: str, *, scopes: Optional[List[str]] = None
    ) -> Response[AccessTokenResponse]:
        r = Route("GET", "/oauth2/token")
        payload = {
            "client_id": self.client_id,
            "client_secret": self._client_secret,
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": self.redirect_uri,
        }
        if scopes:
            payload["scope"] = scopes

        return self.request(r, data=payload)


class AsyncHTTP(BaseHTTP):
    def __init__(
        self,
        connector: Optional[aiohttp.BaseConnector] = None,
        *,
        client_id: int,
        client_secret: str,
        redirect_uri: str,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        http_trace: Optional[aiohttp.TraceConfig] = None,
    ) -> None:
        super().__init__(
            proxy=proxy,
            proxy_auth=proxy_auth,
            client_id=client_id,
            client_secret=client_secret,
            redirect_uri=redirect_uri,
        )
        self.connector: Optional[aiohttp.BaseConnector] = connector
        self.loop: Optional[asyncio.AbstractEventLoop] = loop
        self.http_trace: Optional[aiohttp.TraceConfig] = http_trace
        self.__session: aiohttp.ClientSession = self._create_session()

        user_agent = "OAuth2 (https://github.com/InviteManagerBot/discord.oauth2 {0}) Python/{1[0]}.{1[1]} aiohttp/{2}"
        self.user_agent = user_agent.format(
            __version__, sys.version_info, aiohttp.__version__
        )

    def _create_session(self) -> aiohttp.ClientSession:
        if self.connector is None:
            self.connector = aiohttp.TCPConnector(limit=0)

        return aiohttp.ClientSession(
            connector=self.connector,
            trace_configs=None if self.http_trace is None else [self.http_trace],
            loop=self.loop,
        )

    async def request(self, route: Route, **kwargs: Any) -> Any:
        method = route.method
        url = route.url

        headers = {"User-Agent": self.user_agent, **kwargs.get("headers", {})}

        # discord OAuth2 api requires this Content-Type header
        headers["Content-Type"] = "application/x-www-form-urlencoded"

        if self.proxy:
            kwargs["proxy"] = self.proxy
        if self.proxy_auth:
            kwargs["proxy_auth"] = self.proxy_auth

        # 5 tries
        for trie in range(5):
            async with self.__session.request(method, url, **kwargs) as resp:
                _log.debug("%s %s has returned %s", method, url, resp.status)

                data = await json_or_text(resp)
                status = resp.status

                if 300 > resp.status >= 200:
                    _log.debug("%s %s has received %s", method, url, data)
                    return data

                # rate limit handle
                if status == 429:
                    if not resp.headers.get("Via") or isinstance(data, str):
                        # Banned by Cloudflare more than likely.
                        raise HTTPException(resp, data)

                    retry_after_secs: int = data["retry_after"] / 1000
                    _log.warning(
                        "You are being rate limit, trying again after %.2f seconds",
                        retry_after_secs,
                    )

                    await asyncio.sleep(retry_after_secs)
                    _log.debug("Done sleeping for the rate limit, retrying...")
                    continue

                if status in (500, 502, 504):
                    try_after: int = 1 + trie * 2
                    _log.debug(
                        "Internal error received from %s, trying again after %s seconds",
                        url,
                        try_after,
                    )
                    await asyncio.sleep(try_after)

                if status == 403:
                    raise Forbidden(resp, data)
                elif status == 404:
                    raise NotFound(resp, data)
                elif status >= 500:
                    raise DiscordServerError(resp, data)
                else:
                    raise HTTPException(resp, data)

    async def close(self) -> None:
        if self.__session:
            await self.__session.close()
            self.__session = MISSING



from __future__ import annotations

from typing import TYPE_CHECKING, Optional
from enum import Enum

from .utils import try_enum, parse_time
from .appinfo import IntegrationApp

if TYPE_CHECKING:
    from .types import (
        ServerIntegration as ServerIntegrationPayload,
        IntegrationType,
        IntegrationAccount as IntegrationAccountPayload,
        Snowflake,
        IntegrationApp as IntegrationAppPayload,
        IntegrationAccount as IntegrationAccountPayload,
    )
    from .user import User
    from datetime import datetime

__all__ = ("ExpireBehavior", "IntegrationAccount", "ServerIntegration")


class ExpireBehavior(Enum):
    remove_role = 0
    kick = 1


class IntegrationAccount:
    """Represents integration account.

    Attributes
    ----------
    id: :class:`int`
        Id of the account.
    name: :class:`str`
        Name of the account.
    """

    __slots__ = ("id", "name")

    def __init__(self, data: IntegrationAccountPayload) -> None:
        self.id: int = int(data["id"])
        self.name: str = data["name"]

    def __repr__(self) -> str:
        return f"<IntegrationAccount id={self.id} name={self.name!r}>"


class ServerIntegration:
    """Represents discord server integrations.

    Attributes
    ----------
    id: :class:`int`
        The integration unique id.
    name: :class:`str`
        The integration name.
    type: :class:`IntegrationType`
        The integration type (twitch, youtube, discord, ...).
    enabled: :class:`bool`
        If this integration enabled.
    syncing: :class:`bool`
        If this integration syncing.
    role_id: :class:`int`
        role id that this integration uses for "subscribers".
    enable_emoticons: Optional[:class:`bool`]
        If emoticons should be synced for this integration (twitch only currently).
    expire_behavior: Optional[:class:`ExpireBehavior`]
        The behavior of expiring subscribers (remove_role or kick).
    expire_grace_period: :class:`int`
        The grace period (in days) before expiring subscribers.
    user: :class:`User`
        User for this integration.
    account: :class:`IntegrationAccount`
        Integration account information.
    synced_at: Optional[:class:`datetime.datetime`]
        When this integration was last synced.
    subscriber_count: Optional[:class:`int`]
        How many subscribers this integration has.
    revoked: :class:`bool`
        If has this integration been revoked.
    application: Optional[:class:`IntegrationApp`]
        The bot/OAuth2 application for discord integrations, if available.
    """

    __slots__ = (
        "user",
        "id",
        "name",
        "type",
        "enabled",
        "syncing",
        "role_id",
        "enable_emoticons",
        "expire_behavior",
        "expire_grace_period",
        "account",
        "synced_at",
        "subscriber_count",
        "revoked",
        "application",
    )

    def __init__(self, data: ServerIntegrationPayload, user: User) -> None:
        self.user: User = user
        self.id: int = int(data["id"])
        self.name: str = data["name"]
        self.type: IntegrationType = data["type"]
        self.enabled: bool = data.get("enabled", False)
        self.syncing: bool = data.get("syncing", False)
        self.role_id: Optional[Snowflake] = data.get("role_id")
        self.enable_emoticons: Optional[bool] = data.get("enable_emoticons")
        self.expire_behavior: Optional[ExpireBehavior] = try_enum(
            ExpireBehavior, data.get("expire_behavior")
        )
        self.expire_grace_period: Optional[int] = data.get("expire_grace_period")
        self.account: IntegrationAccount = IntegrationAccount(data["account"])
        self.synced_at: Optional[datetime] = parse_time(data.get("synced_at"))
        self.subscriber_count: Optional[int] = data.get("subscriber_count")
        self.revoked: bool = data.get("revoked", False)
        application: Optional[IntegrationAppPayload] = data.get("application")
        self.application: Optional[IntegrationApp] = (
            IntegrationApp(application) if application else None
        )

    def __repr__(self) -> str:
        return f"<Integration id={self.id} name={self.name!r}>"

    def __str__(self) -> str:
        return self.name



from typing import TypedDict, List
from .integration import ServerIntegration
from .integration import IntegrationType as ConnectionType


class _OptionalConnection(TypedDict, total=False):
    revoked: bool
    integrations: List[ServerIntegration]


class Connection(_OptionalConnection):
    id: int
    name: str
    type: ConnectionType
    verified: bool
    friend_sync: bool
    show_activity: bool
    visibility: int



from typing import TypedDict, Optional
from .snowflake import SnowflakeList
from .user import User


class PartialMember(TypedDict):
    roles: SnowflakeList
    joined_at: str
    deaf: bool
    mute: bool


class Member(PartialMember, total=False):
    avatar: str
    user: User
    nick: str
    premium_since: Optional[str]
    pending: bool
    permissions: str
    communication_disabled_until: str



from typing import TypedDict, Literal, TypeAlias
from .snowflake import Snowflake
from .user import User

IntegrationType: TypeAlias = Literal[
    "youtube",
    "xbox",
    "twitter",
    "twitch",
    "steam",
    "spotify",
    "reddit",
    "playstation",
    "github",
    "facebook",
    "battlenet",
]


class IntegrationAccount(TypedDict):
    id: int
    name: str


class _OptionalIntegration(TypedDict, total=False):
    enabled: bool
    syncing: bool
    role_id: Snowflake
    enable_emoticons: bool
    expire_behavior: int
    expire_grace_period: int
    user: User
    synced_at: str
    subscriber_count: int
    revoked: bool


class ServerIntegration(_OptionalIntegration):
    id: int
    name: str
    type: IntegrationType
    account: IntegrationAccount



from __future__ import annotations

from typing import Any, Callable, Optional, Type, overload, TYPE_CHECKING
from .utils import copy_doc

if TYPE_CHECKING:
    from typing_extensions import Self

__all__ = ("Permissions", "UserFlags")


class BaseFlags:
    value: int

    def __init__(self, value: int) -> None:
        self.value: int = value

    def __repr__(self):
        return f"<{self.__class__.__name__} value={self.value}>"

    def _has_flag(self, other: int) -> bool:
        return (self.value & other) == other


class flag:
    def __init__(self, f: Callable[[Any], int]) -> None:
        self.flag: int = f(None)
        self.__doc__: Optional[str] = f.__doc__

    @overload
    def __get__(self, instance: None, owner: Type[BaseFlags]) -> Self:
        ...

    @overload
    def __get__(self, instance: BaseFlags, owner: Type[BaseFlags]) -> bool:
        ...

    def __get__(self, instance: Optional[BaseFlags], owner: Type[BaseFlags]) -> Any:
        if instance is None:
            return self
        return instance._has_flag(self.flag)

    def __repr__(self):
        return "<flag flag={.flag!r}>".format(self)


flag_alias = flag


class Permissions(BaseFlags):
    __slots__ = ()

    @flag
    def create_instant_invite(self):
        """:class:`bool`: Returns ``True`` if the user can create instant invites."""
        return 1 << 0

    @flag
    def kick_members(self):
        """:class:`bool`: Returns ``True`` if the user can kick users from the guild."""
        return 1 << 1

    @flag
    def ban_members(self):
        """:class:`bool`: Returns ``True`` if a user can ban users from the guild."""
        return 1 << 2

    @flag
    def administrator(self):
        """:class:`bool`: Returns ``True`` if a user is an administrator. This role overrides all other permissions.
        This also bypasses all channel-specific overrides.
        """
        return 1 << 3

    @flag
    def manage_channels(self):
        """:class:`bool`: Returns ``True`` if a user can edit, delete, or create channels in the guild.
        This also corresponds to the "Manage Channel" channel-specific override."""
        return 1 << 4

    @flag
    def manage_guild(self):
        """:class:`bool`: Returns ``True`` if a user can edit guild properties."""
        return 1 << 5

    @flag
    def add_reactions(self):
        """:class:`bool`: Returns ``True`` if a user can add reactions to messages."""
        return 1 << 6

    @flag
    def view_audit_log(self):
        """:class:`bool`: Returns ``True`` if a user can view the guild's audit log."""
        return 1 << 7

    @flag
    def priority_speaker(self):
        """:class:`bool`: Returns ``True`` if a user can be more easily heard while talking."""
        return 1 << 8

    @flag
    def stream(self):
        """:class:`bool`: Returns ``True`` if a user can stream in a voice channel."""
        return 1 << 9

    @flag
    def read_messages(self):
        """:class:`bool`: Returns ``True`` if a user can read messages from all or specific text channels."""
        return 1 << 10

    @flag
    def send_messages(self):
        """:class:`bool`: Returns ``True`` if a user can send messages from all or specific text channels."""
        return 1 << 11

    @flag
    def send_tts_messages(self):
        """:class:`bool`: Returns ``True`` if a user can send TTS messages from all or specific text channels."""
        return 1 << 12

    @flag
    def manage_messages(self):
        """:class:`bool`: Returns ``True`` if a user can delete or pin messages in a text channel.
        .. note::
            Note that there are currently no ways to edit other people's messages.
        """
        return 1 << 13

    @flag
    def embed_links(self):
        """:class:`bool`: Returns ``True`` if a user's messages will automatically be embedded by Discord."""
        return 1 << 14

    @flag
    def attach_files(self):
        """:class:`bool`: Returns ``True`` if a user can send files in their messages."""
        return 1 << 15

    @flag
    def read_message_history(self):
        """:class:`bool`: Returns ``True`` if a user can read a text channel's previous messages."""
        return 1 << 16

    @flag
    def mention_everyone(self):
        """:class:`bool`: Returns ``True`` if a user's @everyone or @here will mention everyone in the text channel."""
        return 1 << 17

    @flag
    def external_emojis(self):
        """:class:`bool`: Returns ``True`` if a user can use emojis from other guilds."""
        return 1 << 18

    @flag
    def view_guild_insights(self):
        """:class:`bool`: Returns ``True`` if a user can view the guild's insights."""
        return 1 << 19

    @flag
    def connect(self):
        """:class:`bool`: Returns ``True`` if a user can connect to a voice channel."""
        return 1 << 20

    @flag
    def speak(self):
        """:class:`bool`: Returns ``True`` if a user can speak in a voice channel."""
        return 1 << 21

    @flag
    def mute_members(self):
        """:class:`bool`: Returns ``True`` if a user can mute other users."""
        return 1 << 22

    @flag
    def deafen_members(self):
        """:class:`bool`: Returns ``True`` if a user can deafen other users."""
        return 1 << 23

    @flag
    def move_members(self):
        """:class:`bool`: Returns ``True`` if a user can move users between other voice channels."""
        return 1 << 24

    @flag
    def use_voice_activation(self):
        """:class:`bool`: Returns ``True`` if a user can use voice activation in voice channels."""
        return 1 << 25

    @flag
    def change_nickname(self):
        """:class:`bool`: Returns ``True`` if a user can change their nickname in the guild."""
        return 1 << 26

    @flag
    def manage_nicknames(self):
        """:class:`bool`: Returns ``True`` if a user can change other user's nickname in the guild."""
        return 1 << 27

    @flag
    def manage_roles(self):
        """:class:`bool`: Returns ``True`` if a user can create or edit roles less than their role's position.
        This also corresponds to the "Manage Permissions" channel-specific override.
        """
        return 1 << 28

    @flag
    def manage_webhooks(self):
        """:class:`bool`: Returns ``True`` if a user can create, edit, or delete webhooks."""
        return 1 << 29

    @flag
    def manage_emojis(self):
        """:class:`bool`: Returns ``True`` if a user can create, edit, or delete emojis."""
        return 1 << 30

    @flag
    def use_slash_commands(self):
        """:class:`bool`: Returns ``True`` if a user can use slash commands."""
        return 1 << 31

    @flag
    def request_to_speak(self):
        """:class:`bool`: Returns ``True`` if a user can request to speak in a stage channel."""
        return 1 << 32


class UserFlags(BaseFlags):
    __slots__ = ()

    @flag
    def staff(self):
        """:class:`bool`: Returns ``True`` if the user is a Discord Employee."""
        return 1 << 0

    @flag
    def partner(self):
        """:class:`bool`: Returns ``True`` if the user is a Discord Partner."""
        return 1 << 1

    @flag
    def hypesquad(self):
        """:class:`bool`: Returns ``True`` if the user is a HypeSquad Events member."""
        return 1 << 2

    @flag
    def bug_hunter_level_1(self):
        """:class:`bool`: Returns ``True`` if the user is a bug hunter level 1."""
        return 1 << 3

    @flag_alias
    @copy_doc(bug_hunter_level_1)
    def bug_hunter(self):
        return 1 << 3

    @flag
    def hypesquad_online_house_1(self):
        """:class:`bool`: Returns ``True`` if the user is a HypeSquad Bravery member."""
        return 1 << 6

    @flag_alias
    @copy_doc(hypesquad_online_house_1)
    def hypesquad_bravery(self):
        return 1 << 6

    @flag
    def hypesquad_online_house_2(self):
        """:class:`bool`: Returns ``True`` if the user is a HypeSquad Brilliance member."""
        return 1 << 7

    @flag_alias
    @copy_doc(hypesquad_online_house_2)
    def hypesquad_brilliance(self):
        return 1 << 7

    @flag
    def hypesquad_online_house_3(self):
        """:class:`bool`: Returns ``True`` if the user is a HypeSquad Balance member."""
        return 1 << 8

    @flag_alias
    @copy_doc(hypesquad_online_house_3)
    def hypesquad_balance(self):
        return 1 << 8

    @flag
    def early_supporter(self):
        """:class:`bool`: Returns ``True`` if the user is an Early Supporter."""
        return 1 << 9

    @flag
    def team_pseudo_user(self):
        """:class:`bool`: Returns ``True`` if the user is a Team User."""
        return 1 << 10

    @flag_alias
    @copy_doc(team_pseudo_user)
    def team_user(self):
        return 1 << 10

    @flag
    def system(self):
        """:class:`bool`: Returns ``True`` if the user is a system user (i.e. represents Discord officially)."""
        return 1 << 12

    @flag
    def bug_hunter_level_2(self):
        """:class:`bool`: Returns ``True`` if the user is a Bug Hunter Level 2"""
        return 1 << 14

    @flag
    def verified_bot(self):
        """:class:`bool`: Returns ``True`` if the user is a Verified Bot."""
        return 1 << 16

    @flag
    def verified_developer(self):
        """:class:`bool`: Returns ``True`` if the user is an Early Verified Bot Developer."""
        return 1 << 17

    @flag
    def certified_moderator(self):
        """:class:`bool`: Returns ``True`` if the user is a Discord Certified Moderator."""
        return 1 << 18

    @flag
    def bot_http_interactions(self):
        """:class:`bool`: Returns ``True`` if the user is a bot that only uses HTTP interactions
        and is shown in the online member list."""
        return 1 << 19

    @flag
    def spammer(self):
        """:class:`bool`: Returns ``True`` if the user is flagged as a spammer by Discord."""
        return 1 << 20



from typing import TypedDict, Optional
from .user import User
from .snowflake import Snowflake


class IntegrationAccount(TypedDict):
    id: Snowflake
    name: str


class _OptionalIntegrationApp(TypedDict, total=False):
    bot: User


class IntegrationApp(_OptionalIntegrationApp):
    id: Snowflake
    name: str
    icon: Optional[str]
    description: str



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union, Dict

if TYPE_CHECKING:
    from aiohttp import ClientResponse

    Response = ClientResponse

__all__ = (
    "Oauth2Exception",
    "HTTPException",
    "Forbidden",
    "NotFound",
    "DiscordServerError",
)


class Oauth2Exception(Exception):
    """Base exception class for discord oauth2"""

    pass


class HTTPException(Oauth2Exception):
    """Exception that's raised when HTTP request fails.

    Attributes
    ---------
    response: Union[:class:`requests.Response`, :class:`aiohttp.ClientResponse`]
        The response object of the failed HTTP request.
        This will be an :class:`aiohttp.ClientResponse`
        instance if you are using AsyncClient, otherwise
        it will be :class:`requests.Session`.
    message: :class:`str`
        The message of the error.
        Could be an empty string.
    status: :class:`int`
        The status code of the HTTP request.
        e.g. 404, 403, 500
    """

    def __init__(self, response: Response, message: Union[str, Dict[str, Any]]) -> None:
        self.response: Response = response
        self.status = response.status

        if isinstance(message, dict):
            self.code: int = message.get("code", 0)
            self.message: str = message.get("error_description", message.get("message"))
        else:
            self.code: int = 0
            self.message: str = message

        fmt = "{0.status} {0.reason} (error code: {1})"
        super().__init__(fmt.format(response, self.code))


class Forbidden(HTTPException):
    """Exception that's raised for when status code 403.

    Subclass of :exc:`HTTPException`
    """

    pass


class NotFound(HTTPException):
    """Exception that's raised for when status code 404.

    Subclass of :exc:`HTTPException`
    """

    pass


class DiscordServerError(HTTPException):
    """Exception that's raised for when a 500 range status code.

    Subclass of :exc:`HTTPException`
    """

    pass



import aiohttp
import asyncio

from weakref import WeakValueDictionary
from typing import List, Optional

from .http import AsyncHTTP
from .token import AccessToken
from .user import User, PartialUser

__all__ = ("Client",)


class Client:
    """Represents a client that offers interaction to the discord OAuth2 API.

    Parameters
    -----------
    client_id: :class:`int`
        The OAuth2 application client id.
    client_secret: :class:`str`
        The OAuth2 application client secret.
    redirect_uri: :class:`str`
        The OAuth2 redirect url.
        This url must configurated from uri's at discord developer portal.
    scopes: Optional[List[:class:`str`]]
        List of OAuth2 scopes.
        Default ``None``.
    connector: Optional[:class:`aiohttp.BaseConnector`]
        A connector for aiohttp client API.
        `aiohttp documentation <https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.BaseConnector>`.
        Default ``None``.
    loop: Optional[:class:`asyncio.AbstractEventLoop`]
        The event loop of asyncio.
        This loop will be only used at the instance of :class:`aiohttp.ClientSession`.
        Default ``None``.
    proxy: Optional[:class:`str`]
        The proxy URL.
        Default ``None``.
    proxy_auth: Optional[:class:`aiohttp.BasicAuth`]
        Object that represents proxy HTTP Basic Authorization.
        `aiohttp documentation <https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.BasicAuth>`.
        Default ``None``.
    http_trace: Optional[:class:`aiohttp.TraceConfig`]
        The trace configuration to use for tracking HTTP requests the library does using ``aiohttp``.
        `aiohttp documentation <https://docs.aiohttp.org/en/stable/client_advanced.html#client-tracing>`.
        Default ``None``.
    """

    def __init__(
        self,
        *,
        client_id: int,
        client_secret: str,
        redirect_uri: str,
        scopes: Optional[List[str]] = None,
        connector: Optional[aiohttp.BaseConnector] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        http_trace: Optional[aiohttp.TraceConfig] = None,
    ) -> None:
        self.client_id: int = client_id
        self._client_secret: str = client_secret
        self.redirect_uri: str = redirect_uri
        self.scopes: Optional[List[str]] = scopes
        self.http: AsyncHTTP = AsyncHTTP(
            client_id=client_id,
            client_secret=client_secret,
            redirect_uri=redirect_uri,
            connector=connector,
            loop=loop,
            proxy=proxy,
            proxy_auth=proxy_auth,
            http_trace=http_trace,
        )

        self._users: WeakValueDictionary[int, User] = WeakValueDictionary()

    async def exchange_code(self, code: str) -> AccessToken:
        """Exchanges the code you receive from the OAuth2 redirect.

        Parameters
        ---------
        code: :class:`str`
            The code from the querystring.

        Returns
        -------
        :class:`AccessToken`
        """
        data = await self.http.exchange_code(code, scopes=self.scopes)
        return AccessToken(data)

    async def refresh_token(self, refresh_token: str) -> AccessToken:
        """Refreshes an access token.
        This can be useful to anticipate the expiration and refresh the token.

        Parameters
        ---------
        refresh_toke: :class:`str`
            The user's refresh token.

        Returns
        -------
        :class:`AccessToken`
        """
        data = await self.http.refresh_token(refresh_token)
        return AccessToken(data)

    async def fetch_user(self, access_token: AccessToken) -> User:
        """Fetch user's information from discord api using user's access token.

        You must have the scope `identify` to use this.
        To access user's email you must have the scope `email`.

        Parameters
        ----------
        access_token: :class:`AccessToken`
            The user's access token.

        Returns
        ------
        :class:`User`
        """
        data = await self.http.get_user(access_token.access_token)
        user = User(http=self.http, access_token=access_token, data=data)
        self._users[user.id] = user
        return user

    def get_user(self, id: int, /) -> Optional[User]:
        """Returns a user with the given ID

        Parameters
        ----------
        id: :class:`int`
            The user's id that you
            want to find.

        Returns
        -------
        Optional[:class:`User`]
            Returns ``None`` when there's no user with that id
            in the cache
        """
        return self._users.get(id)

    def get_partial_user(self, *, access_token: AccessToken) -> PartialUser:
        """Returns a partial user.

        This is useful if you have a access_token but don't want to do an API call
        to fetch it again.

        Note that this partial user class will not let you use methods that requires
        user information.

        Parameters
        ----------
        access_token: :class:`AccessToken`
            The user's access token.

        Returns
        ------
        :class:`PartialUser`
        """
        return PartialUser(http=self.http, access_token=access_token)

    async def close(self):
        """Closes and cleanup operations on the client."""
        await self.http.close()



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, List

from .guild import Guild
from .connection import Connection
from .flags import UserFlags

if TYPE_CHECKING:
    from .types.user import PartialUser as UserPayload
    from .http import AsyncHTTP
    from .token import AccessToken
    from .types import Snowflake


__all__ = ("BaseUser", "User", "PartialUser")


class BaseUser:
    """Represents your Discord user that contains information from the OAuth2 API.

    .. container:: operations

        ..  describe:: x == y

            Compare the ids of the two users and verify that they are the same.

        .. describe:: x != y

            Compare the ids of the two users and verify that they are the different.

        .. describe:: hash(x)

            Return the user's hash.

        .. describe:: str(x)

            Returns the user's id with discriminator.

    Attributes
    -----------
    id: :class:`int`
        The user's unique id.
    name: :class:`str`
        The user's username.
    discriminator: :class:`str`
        The user's discriminator.
    email: Optional[:class:`str`]
        The user's email.
    bot: :class:`bool`
        If the current user is a bot account.
    system: :class:`bool`
        If the user is a system user.
    verified: :class:`bool`
        If the user's email is verified.
    locale: Optional[:class:`str`]
        The IETF language tag used to identify the language the user is using.
    mfa_enabled: :class:`bool`
        Specifies if the user has MFA turned on.
    """

    __slots__ = (
        "id",
        "name",
        "discriminator",
        "email",
        "mfa_enabled",
        "locale",
        "verified",
        "bot",
        "system",
        "_http",
        "_access_token",
        "_avatar_hash",
        "_banner_hash",
        "_accent_colour",
        "_public_flags",
        "_flags",
        "guilds",
        "connections",
    )

    def __init__(
        self, *, http: AsyncHTTP, access_token: AccessToken, data: UserPayload
    ) -> None:
        self._http: AsyncHTTP = http
        self._access_token: AccessToken = access_token
        self._update(data)

        self.guilds: List[Guild] = []
        self.connections: List[Connection] = []

    def _update(self, data: UserPayload):
        self.id: int = int(data["id"])
        self.name: str = data["username"]
        self.discriminator: int = int(data["discriminator"])
        self.email: Optional[str] = data.get("email", None)
        self._avatar_hash: Optional[str] = data["avatar"]
        self._banner_hash: Optional[str] = data.get("banner", None)
        self._accent_colour: Optional[int] = data.get("accent_color", None)
        self._public_flags: int = data.get("public_flags", 0)
        self._flags: int = data.get("flags", 0)
        self.mfa_enabled: bool = data.get("mfa_enabled", False)
        self.locale: Optional[str] = data.get("locale")
        self.verified: bool = data.get("verified", False)
        self.bot: bool = data.get("bot", False)
        self.system: bool = data.get("system", False)

    def __repr__(self) -> str:
        return f"<User id={self.id} name={self.name!r} discriminator={self.discriminator!r} bot={self.bot}>"

    def __str__(self) -> str:
        return f"{self.id}#{self.discriminator}"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, BaseUser) and other.id == self.id

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return self.id >> 22

    @property
    def flags(self) -> UserFlags:
        """:class:`Permissions`: Returns the resolved flags of the user."""
        return UserFlags(self._flags)

    @property
    def public_flags(self) -> UserFlags:
        """:class:`Permissions`: Returns the resolved public flags of the user."""
        return UserFlags(self._public_flags)

    @property
    def avatar_url(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns the link of cdn for the avatar the user has."""
        if self._avatar_hash is None:
            return None

        animated = self._avatar_hash.startswith("a_")
        fmt = "gif" if animated else "png"
        return f"https://cdn.discordapp.com/avatars/{self.id}/{self._avatar_hash}.{fmt}?size=1024"

    @property
    def default_avatar(self) -> str:
        """:class:`str`: Returns the default avatar for a given user.
        This is calculated by the user's discriminator.
        """
        index: int = int(self.discriminator) % 5
        return f"https://cdn.discordapp.com/embed/avatars/{index}.png"

    @property
    def display_avatar(self) -> str:
        """:class:`str`: Returns the user's display avatar.

        This method is similar to avatar_url, but
        if the user doesn't have an avatar it will give the default avatar.
        """
        return self.avatar_url or self.default_avatar

    @property
    def banner(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns the user's banner asset, if available."""
        if self._banner_hash is None:
            return None

        animated = self._banner_hash.startswith("a_")
        fmt = "gif" if animated else "png"
        return f"https://cdn.discordapp.com/banners/{self.id}/{self._banner_hash}.{fmt}?size=512"

    @property
    def accent_colour(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns the user's accent colour, if applicable.

        .. note::
            This property will return the color in hex format.
        """
        return f"#{self._accent_colour:0>6x}"

    @property
    def mention(self) -> str:
        """:class:`str`: Returns a string format to mention this user in discord."""
        return f"<@{self.id}>"


class User(BaseUser):
    __slots__ = ("__weakref__",)

    async def refresh(self) -> AccessToken:
        """Refreshes the access token for the user.

        Returns
        -------
        :class:`AccessToken`
        """
        access_token = self._access_token.access_token
        data = await self._http.refresh_token(refresh_token=access_token)
        self._access_token = AccessToken(data)
        return self._access_token

    async def add_guild(
        self,
        guild: Guild,
        bot_token: str,
        *,
        nick: Optional[str] = None,
        roles: Optional[List[Snowflake]] = None,
        mute: Optional[bool] = None,
        deaf: Optional[bool] = None,
    ) -> None:
        """Adds a user to the guild.

        You must have the scope `guilds.join` to use this.

        Parameters
        ---------
        guild: :class:`.Guild`
            Guild that the user will join.
        bot_token: :class:`str`
            A bot token that is in the target guild.
            Discord requires a bot that you have a bot
            that is in the target guild you want to add the user.
        nick: Optional[:class:`str`]
            Value to set user's nickname to.
        roles: Optional[List[:class:`int`]]
            A list of role ids that will be added
            to the member.
        mute: Optional[:class:`bool`]
            Whether the user is muted in voice channels
        deaf: Optional[:class:`bool`]
            Whether the user is deafened in voice channels
        """
        await self._http.add_user_guild(
            self.id,
            guild.id,
            self._access_token.access_token,
            bot_token,
            nick=nick,
            roles=roles,
            mute=mute,
            deaf=deaf,
        )

    async def fetch_guilds(self) -> List[Guild]:
        """Fetch a list of partial guilds that the user is a member of.

        You must have the scope `guilds` to use this.

        Returns
        -------
        List[:class:`.Guild`]
        """
        data = await self._http.get_user_guilds(self._access_token.access_token)
        self.guilds = [Guild(data=x, user=self, http=self._http) for x in data]
        return self.guilds

    async def fetch_connections(self) -> List[Connection]:
        """Fetch a list of connections of the user.

        You must have the scope `connections` to use this.

        Returns
        -------
        List[:class:`Connection`]
        """
        data = await self._http.get_user_connections(self._access_token.access_token)
        self.connections = [Connection(x, self) for x in data]
        return self.connections


class PartialUser(User):
    """Represents a partial user

    Note that this class is trimmed down and has no rich attributes.

    .. container:: operations

        ..  describe:: x == y

            Compare the ids of the two users and verify that they are the same.

        .. describe:: x != y

            Compare the ids of the two users and verify that they are the different.

        .. describe:: hash(x)

            Return the user's hash.
    """

    __slots__ = ("_http", "_access_token", "id")

    def __init__(self, *, http: AsyncHTTP, access_token: AccessToken) -> None:
        self._http = http
        self._access_token = access_token
        self.id = 0

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id}>"

    async def add_guild(self, *args, **kwargs):
        # This method requires user id
        raise NotImplementedError



from __future__ import annotations

from typing import Optional, List, TYPE_CHECKING

from .member import Member
from .flags import Permissions

if TYPE_CHECKING:
    from .types import PartialGuild as PartialGuildPayload
    from .http import AsyncHTTP
    from .user import User

__all__ = ("Guild",)


class Guild:
    """Represents a Discord partial guild.

    This is referred to as a "server" in the official Discord UI.

    .. container:: operations

        .. describe:: x == y

            Checks if two guilds are equal.

        .. describe:: x != y

            Checks if two guilds are not equal.

        .. describe:: hash(x)

            Returns the guild's hash.

        .. describe:: str(x)

            Returns the guild's name.

    Attributes
    ----------
    id: :class:`int`
        The guild id.
    name: :class:`str`
        The guild name.
    permissions: Optional[:class:`Permissions`]
        An Permissions class that represents
        the permissions of the user in the guild.
    features: Optional[List[:class:`str`]]
        A list of features that the guild has.
    """

    def __init__(self, data: PartialGuildPayload, user: User, http: AsyncHTTP) -> None:
        self._http: AsyncHTTP = http
        self._user: User = user
        self.id: int = int(data["id"])
        self.name: str = data["name"]
        self._icon_hash: Optional[str] = data.get("icon")
        self._is_owner: Optional[bool] = data.get("owner")
        self._permissions: int = int(data.get("permissions", 0))
        self.features: Optional[List[str]] = data.get("features")

    def __repr__(self) -> str:
        return f"<Guild id={self.id} name={self.name!r}>"

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Guild) and other.id == self.id

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return self.id >> 22

    @property
    def permissions(self) -> Permissions:
        """:class:`Permissions`: Returns the resolved permissions that the oauth2 user has in this guild."""
        return Permissions(self._permissions)

    @property
    def icon(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns the guild's icon url, if available."""
        if not self._icon_hash:
            return None

        animated = self._icon_hash.startswith("a_")
        fmt = "gif" if animated else "png"
        return f"https://cdn.discordapp.com/icons/{self.id}/{self._icon_hash}.{fmt}?size=1024"

    @property
    def is_owner(self) -> Optional[bool]:
        """Optional[:class:`str`]: Returns if the current user is the owner of the guild."""
        return self._is_owner

    async def fetch_me(self) -> Member:
        """Fetch the current user member information of this guild.

        You must have the scope `guilds.members.read` to use this.

        Returns
        -------
        :class:`Member`
        """
        data = await self._http.get_member(
            self._user._access_token.access_token, self.id
        )
        return Member(data, user=self._user, guild=self)
